/*************************************************************************\
* Copyright (c) 2006 The University of Chicago, as Operator of Argonne
*     National Laboratory.
* This software is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
\*************************************************************************/

/*
 * powrboot device support
 */

#include <epicsStdio.h>
#include <ctype.h>
#include <devCommonGpib.h>

/******************************************************************************
 *
 * The following define statements are used to declare the names to be used
 * for the dset tables.   
 *
 * A DSET_AI entry must be declared here and referenced in an application
 * database description file even if the device provides no AI records.
 *
 ******************************************************************************/
#define DSET_AI     devAipowrboot
#define DSET_BO     devBopowrboot
#define DSET_LI     devLipowrboot
#define DSET_LO     devLopowrboot

#include <devGpib.h> /* must be included after DSET defines */

#define TIMEOUT     12.0   /* I/O must complete within this time */
#define TIMEWINDOW  2.0    /* Wait this long after device timeout */

/******************************************************************************
 * Strings used by the init routines to fill in the znam,onam,...
 * fields in BI and BO record types.
 ******************************************************************************/

static char  *offOnNames[] = { "Off","On" };
static struct devGpibNames   offOn = { 2,offOnNames,0,1 };

/******************************************************************************
 * String arrays for EFAST operations. The last entry must be 0.
 *
 * On input operations,only as many bytes as are found in the string array
 * elements are compared.  Additional bytes are ignored.
 * The first matching string  will be used as a match.
 *
 * For the input operations,the strings are compared literally!  This
 * can cause problems if the instrument is returning things like \r and \n
 * characters.  When defining input strings so you include them as well.
 ******************************************************************************/

static char *offOnEfast[] = {"OFF\r","ON\r",0};

/*
 * Custom conversion routine
 * What a truly horrible communication protocol.
 * We look for lines of the form:
 *                  <w> <plug number> <anything> <OFF|ON> <w>
 * where <w> indicates arbitrary white space.
 */
static int
convertStatus(struct gpibDpvt *pdpvt, int P1, int P2, char **P3)
{
    struct longinRecord *pli = ((struct longinRecord *)(pdpvt->precord));
    const char *cp, *lp, *ep;
    int plugNum = 1;
    int val = 0;

    cp = pdpvt->msg;
    ep = cp + pdpvt->msgInputLen;
    lp = cp;
    for(;;) {
        if (cp >= ep) {
            epicsSnprintf(pdpvt->pasynUser->errorMessage,
                      pdpvt->pasynUser->errorMessageSize,
                      "Truncated reply");
            return -1;
        }
        if (*cp == '\n') {
            asynPrintIO(pdpvt->pasynUser, ASYN_TRACEIO_FILTER, lp, (cp-lp) + 1,
                                        "Checking line: ");
            /*
             * Look for a line beginning with the plug number.
             */
            while (isspace(*lp) && (lp != cp))
                lp++;
            if (*lp == ('0' + plugNum)) {
                asynPrint(pdpvt->pasynUser, ASYN_TRACEIO_FILTER,
                                                    "Match plug %d\n", plugNum);
                /*
                 * Look for ON or OFF at the end of the line.
                 */
                lp = cp;
                while (isspace(*lp))
                    lp--;
                if ((lp[0] == 'N') && (lp[-1] == 'O')) {
                    val |= 1 << (plugNum - 1);
                }
                else if ((lp[0] != 'F') || (lp[-1] != 'F') || (lp[-2] != 'O')) {
                    epicsSnprintf(pdpvt->pasynUser->errorMessage,
                              pdpvt->pasynUser->errorMessageSize,
                              "Garbled reply");
                    return -1;
                }
                if (++plugNum > 5)
                    break;
            }
            lp = cp;
        }
        cp++;
    }
    asynPrint(pdpvt->pasynUser, ASYN_TRACEIO_FILTER, "Val: %#x\n", val);
    pli->val = val;
    return 0;
}

/******************************************************************************
 * Array of structures that define all GPIB messages
 * supported for this type of instrument.
 ******************************************************************************/
#define OFFON(N) {&DSET_BO,GPIBEFASTO,IB_Q_HIGH,"/"#N" ",NULL,1500,20,NULL,0,0,offOnEfast,&offOn,"+>"}
#define BOOT(N) {&DSET_BO,GPIBWRITE,IB_Q_HIGH,NULL,"/"#N" BOOT\r",1500,20,NULL,0,0,NULL,NULL,"+>"}
static struct gpibCmd gpibCmds[] = {
    /* Command 0: Soak up initial message/prompt */
    {&DSET_AI,GPIBREAD,IB_Q_HIGH,"","",0,1500,NULL,0,0,NULL,NULL,"+>"},

    /* Commands 1-5: Output 1-5 OFF/ON */
    OFFON(1),
    OFFON(2),
    OFFON(3),
    OFFON(4),
    OFFON(5),

    /* Commands 6-10: Output 1-5 BOOT */
    BOOT(1),
    BOOT(2),
    BOOT(3),
    BOOT(4),
    BOOT(5),

    /* Command 11: Get off/on status */
    {&DSET_LI,GPIBREAD,IB_Q_HIGH,"/S\r",0,0,1500,convertStatus,0,0,NULL,NULL,"+>"},
};

/* The following is the number of elements in the command array above.  */
#define NUMPARAMS sizeof(gpibCmds)/sizeof(struct gpibCmd)

/******************************************************************************
 * Initialize device support parameters
 *
 *****************************************************************************/
static long init_ai(int parm)
{
    if(parm==0) {
        devSupParms.name = "devpowrboot";
        devSupParms.gpibCmds = gpibCmds;
        devSupParms.numparams = NUMPARAMS;
        devSupParms.timeout = TIMEOUT;
        devSupParms.timeWindow = TIMEWINDOW;
        devSupParms.respond2Writes = 0;
    }
    return(0);
}
